"""Data block representing an integer."""

from ._abstract_block import AbstractDataBlock


class IntegerBlock(AbstractDataBlock):
    """Integer data block."""

    _DEF = [
        (0x01, 1),
        (0x02, 2),
        (0x03, 3),
        (0x04, 4),
        (0x06, 6),
        (0x07, 8),
    ]

    _DIF = {l: b for (b, l) in _DEF}
    _LENGTH = {b: l for (b, l) in _DEF}

    DATA_FIELD_SPECIFIERS = [b for b, _ in _DEF]

    def __init__(self, value=0, length=1):
        """Initialize integer data block."""
        super().__init__()

        self.value = value
        self.length = length

    def __len__(self):
        """Return length of data field."""
        return self.length

    @property
    def dif(self):
        """Return DIF."""
        return (self._dif & 0xF0) | self._DIF[len(self)]

    @property
    def data(self):
        """Get byte representation of value."""
        return int.to_bytes(
            self.value, length=len(self), byteorder="little", signed=True
        )

    @classmethod
    def from_stream(cls, dif, stream):
        """Construct IntegerBlock from stream."""
        # Create instance and parse DRH
        block = super().from_stream(dif, stream)

        block.length = cls._LENGTH[dif & 0x0F]

        data = stream.read(len(block))
        block.value = int.from_bytes(data, byteorder="little", signed=True)

        return block



"""Abstract MBus frame."""
import binascii


class AbstractMBusFrame:
    """Abstract MBus frame."""

    def __repr__(self):
        """Return a readable byte representation of the MBus frame."""
        _bytes = binascii.hexlify(bytes(self), sep=" ")
        return f"<{self.__class__.__name__} {_bytes}>"

    def __eq__(self, other: object) -> bool:
        """Test for bytewise equality."""
        return bytes(self) == bytes(other)



"""MBus frame class."""

import enum
import struct

from .._common import Byte
from ..data import VariableData
from ._abstract_frame import AbstractMBusFrame
from ._raw_frame import RawMBusFrame


class FrameType(enum.Enum):
    """Valid frame types."""

    ACKNOWLEDGE = 1
    SHORT = 2
    LONG = 3


class Control(Byte):  # pylint: disable=too-few-public-methods
    """Valid control byte values."""

    SND_NKE = 0x40, 0b11011111
    SND_UD = 0x53, 0b11011111
    REQ_UD2 = 0x5B, 0b11011111
    REQ_UD1 = 0x5A, 0b11011111
    RSP_UD = 0x08, 0b11001111


class ControlInformation(Byte):  # pylint: disable=too-few-public-methods
    """Valid control information byte values."""

    DATA_SEND = 0x51, 0b11111011
    SLAVE_SELECT = 0x52, 0b11111011

    APPLICATION_RESET = 0x50, 0xFF
    SYNC_ACTION = 0x54, 0xFF

    GENERAL_APPLICATION_ERROR = 0x70, 0xFF
    ALARM_STATUS = 0x71, 0xFF

    VARIABLE_DATA = 0x72, 0b11111011
    FIXED_DATA = 0x73, 0b11111011


FCB_BIT = 5

ACD_BIT = 5
DFC_BIT = 4


class Frame(AbstractMBusFrame):
    """MBus frame."""

    def __init__(
        self,
        frame_type: FrameType,
        control: Control = None,
        address: int = 0xFF,
        control_information: ControlInformation = None,
        **kwargs,
    ):
        """Initialize MBus frame."""
        self._type = frame_type

        self._c = control.value if control is not None else None
        self._a = address
        self._ci = control_information

        if self._c is not None:
            if Control(self._c) in [
                Control.SND_NKE,
                Control.SND_UD,
                Control.REQ_UD2,
                Control.REQ_UD1,
            ]:
                self._c |= kwargs.get("fcb", False) << FCB_BIT

            if Control(self._c) in [
                Control.RSP_UD,
            ]:
                self._c |= kwargs.get("acd", False) << ACD_BIT
                self._c |= kwargs.get("dfc", False) << DFC_BIT

        self.data = b""

    def flip_fcb(self):
        """Flip the frame count bit."""
        self._c ^= 1 << FCB_BIT

    def __bytes__(self):  # pylint: disable=inconsistent-return-statements
        """Return byte representation of MBus frame."""
        if self._type == FrameType.ACKNOWLEDGE:
            return b"\xE5"

        if self._type == FrameType.SHORT:
            return b"\x10%b%c\x16" % (self.body, self.checksum)

        if self._type == FrameType.LONG:
            return b"\x68%c%c\x68%b%c\x16" % (
                len(self),
                len(self),
                self.body,
                self.checksum,
            )

    @property
    def body(self):
        """Return data body of MBus frame."""
        if self._type == FrameType.SHORT:
            return b"%c%c" % (self._c, self._a)

        if self._type == FrameType.LONG:
            return b"%c%c%c%b" % (
                self._c,
                self._a,
                self._ci,
                self.data,
            )

        return b""

    def __len__(self):
        """Return length of MBus frame data as indicated in the header."""
        return len(self.body)

    @property
    def checksum(self):
        """Return checksum."""
        return sum([int(byte) for byte in self.body]) % 256

    @classmethod
    def from_raw_frame(cls, raw_frame: RawMBusFrame):
        """Create MBus frame from byte reader."""
        if raw_frame.first_byte == b"\xE5":
            return cls(frame_type=FrameType.ACKNOWLEDGE)

        if raw_frame.first_byte == b"\x10":
            frame = cls(frame_type=FrameType.SHORT)
            frame._c, frame._a = struct.unpack("BB", raw_frame.body)

            return frame

        if raw_frame.first_byte == b"\x68":
            frame = cls(frame_type=FrameType.LONG)
            frame._c, frame._a, frame._ci = struct.unpack("BBB", raw_frame.body[:3])

            if len(raw_frame.body) > 3:
                _ci = ControlInformation(frame._ci)
                if _ci == ControlInformation.VARIABLE_DATA:
                    data_cls = VariableData
                else:
                    raise NotImplementedError

                frame.data = data_cls.from_bytes(raw_frame.body[3:])

            return frame

        raise ValueError(f"Unknown first byte {raw_frame.first_byte:b}")



"""Data blocks."""

from ._abstract_block import AbstractBlock, AbstractDataBlock
from ._bcd_block import BCDBlock
from ._blocks import NoDataBlock, SelectionForReadoutBlock
from ._int_block import IntegerBlock
from ._lvar_block import VariableLengthBlock
from ._real_block import RealBlock
from ._special_functions_block import Function, SpecialFunctionsBlock

_CLS = [
    BCDBlock,
    NoDataBlock,
    SelectionForReadoutBlock,
    IntegerBlock,
    VariableLengthBlock,
    RealBlock,
    SpecialFunctionsBlock,
]

BLOCK_TYPES = {_s: cls for cls in _CLS for _s in cls.DATA_FIELD_SPECIFIERS}

__all__ = [
    "AbstractBlock",
    "AbstractDataBlock",
    "BCDBlock",
    "NoDataBlock",
    "SelectionForReadoutBlock",
    "IntegerBlock",
    "VariableLengthBlock",
    "RealBlock",
    "SpecialFunctionsBlock",
    "Function",
]



"""Variable MBus data."""

import io
from fcntl import F_UNLCK

from .block import BLOCK_TYPES, AbstractBlock, Function, SpecialFunctionsBlock


class VariableData:
    """Variable MBus data."""

    def __init__(self):
        """Initialize variable data object."""
        self._blocks = []

        self.header = b""
        self.manufacturer_specific = b""

    def __bytes__(self):
        """Return byte representation."""
        _blocks = b"".join([bytes(block) for block in self._blocks])

        return b"%b%b%b" % (
            self.header,
            _blocks,
            self.manufacturer_specific,
        )

    def append_block(self, block: AbstractBlock):
        """Add block."""
        self._blocks.append(block)

    def _parse_header(self, header: bytes):
        """Parse 12 byte header."""
        # TODO: parse header
        self.header = header

    @property
    def more_frames_to_follow(self):
        """Check if more frames are to follow."""
        return (
            isinstance(self._blocks[-1], SpecialFunctionsBlock)
            and self._blocks[-1].function == Function.NEXT_TELEGRAM
        )

    @classmethod
    def from_bytes(cls, data: bytes):
        """Parse data object from byte stream."""
        data_obj = cls()

        # Convert data into BytesIO object to consume it piece by piece
        stream = io.BytesIO(data)

        # The first 12 bytes contain a fixed data header
        _header = stream.read(12)
        data_obj._parse_header(_header)

        # We begin to parse the data blocks
        while True:
            # Read next data block
            _dif = ord(stream.read(1))
            block_cls = BLOCK_TYPES[_dif & 0x0F]

            block = block_cls.from_stream(_dif, stream)
            data_obj.append_block(block)

            if block_cls == SpecialFunctionsBlock and (
                block.function == Function.END_OF_USER_DATA
                or block.function == Function.NEXT_TELEGRAM
            ):
                break

        data_obj.manufacturer_specific = stream.read()

        return data_obj



"""Variable length data block."""

from ._abstract_block import AbstractDataBlock
from ._bcd_block import decode_bcd, encode_bcd


class VariableLengthBlock(AbstractDataBlock):
    """Variable length data block."""

    DATA_FIELD_SPECIFIERS = [
        0x0D,
    ]

    def __init__(self, lvar=0, value=0):
        """Initialize variable length data block."""
        super().__init__()

        self._dif = 0x0D

        self.lvar = lvar
        self.value = 0

    @classmethod
    def from_stream(cls, dif, stream):
        """Construct VariableLengthBlock from stream."""
        block = super().from_stream(dif, stream)

        block.lvar = ord(stream.read(1))
        block.raw_value = stream.read(len(block))
        return block

    @property
    def raw_value(self):
        """Encode value to bytes."""
        if 0x00 <= self.lvar <= 0xBF:
            # ASCII
            return bytes(reversed(self.value))

        if 0xC0 <= self.lvar <= 0xCF:
            # Positive BCD number
            return encode_bcd(self.value, length=len(self))

        if 0xD0 <= self.lvar <= 0xDF:
            # Negative BCD number
            return encode_bcd(-self.value, length=len(self))

        if 0xE0 <= self.lvar <= 0xEF:
            # Binary number
            int.to_bytes(self.value, length=len(self), byteorder="little", signed=True)

        raise NotImplementedError(f"No encoder for LVAR {self.lvar:x}")

    @raw_value.setter
    def raw_value(self, raw_data):
        """Decode raw data to value."""
        if 0x00 <= self.lvar <= 0xBF:
            # ASCII
            self.value = bytes(reversed(raw_data))
        elif 0xC0 <= self.lvar <= 0xCF:
            # Positive BCD number
            self.value = decode_bcd(raw_data)
        elif 0xD0 <= self.lvar <= 0xDF:
            # Negative BCD number
            self.value = -decode_bcd(raw_data)
        elif 0xE0 <= self.lvar <= 0xEF:
            # Binary number with
            self.value = int.from_bytes(raw_data, byteorder="little", signed=True)
        else:
            raise NotImplementedError(f"No decoder for LVAR {self.lvar:x}")

    def __len__(self):
        """Length of data."""
        # We have a variable length block and have to calculate the length of the data
        if 0x00 <= self.lvar <= 0xBF:
            return self.lvar

        if 0xC0 <= self.lvar <= 0xCF:
            return self.lvar - 0xC0

        if 0xD0 <= self.lvar <= 0xDF:
            return self.lvar - 0xD0

        if 0xE0 <= self.lvar <= 0xEF:
            return self.lvar - 0xE0

        if 0xF0 <= self.lvar <= 0xFA:
            return self.lvar - 0xF0

        raise ValueError(f"LVAR field {self.lvar:x} not supported")

    @property
    def data(self):
        """Return byte representation of data."""
        return b"%c%b" % (self.lvar, self.raw_value)



"""Raw data MBus frame."""

import asyncio
import struct

from ._abstract_frame import AbstractMBusFrame
from ._util import calculate_checksum


class RawMBusFrame(AbstractMBusFrame):
    """Raw MBus frame as received via a serial connection."""

    def __init__(self):
        """Initialize MBus frame."""
        self.first_byte = b""
        self.head = b""
        self.data_head = b""
        self.data = b""
        self.foot = b""

    @property
    def body(self):
        """Return byte representation of the frame body."""
        return b"%b%b" % (self.data_head, self.data)

    def __bytes__(self):
        """Return byte representation of the frame."""
        return b"%b%b%b%b" % (self.first_byte, self.head, self.body, self.foot)

    @classmethod
    async def from_reader(cls, reader: asyncio.StreamReader):
        """Create raw MBus frame from stream reader."""
        # Read the first byte
        frame = cls()
        frame.first_byte = await reader.readexactly(1)

        if frame.first_byte == b"\xE5":
            return frame

        if frame.first_byte == b"\x10":
            frame.data_head = await reader.readexactly(2)

            frame.foot = _buf = await reader.readexactly(2)
            checksum, stop = struct.unpack("BB", _buf)

            assert calculate_checksum(frame.body) == checksum, "Checksum mismatch"
            assert stop == 0x16, "Invalid stop byte"

            return frame

        if frame.first_byte == b"\x68":
            frame.head = _buf = await reader.readexactly(3)

            length1, length2, start = struct.unpack("BBB", _buf)

            assert length1 == length2, "Length doesn't match"
            assert start == 0x68, "Unexpected start byte"

            frame.data_head = await reader.readexactly(3)

            data_len = length1 - 3
            if data_len > 0:
                frame.data = await reader.readexactly(data_len)

            frame.foot = _buf = await reader.readexactly(2)
            checksum, stop = struct.unpack("BB", _buf)

            assert calculate_checksum(frame.body) == checksum, "Checksum mismatch"
            assert stop == 0x16, "Invalid stop byte"

            return frame

        raise ValueError(f"Unknown first byte {frame.first_byte:b}")



"""Special function block."""
from enum import Enum

from ._abstract_block import AbstractBlock


class Function(Enum):
    """Special function."""

    END_OF_USER_DATA = 0x0F
    NEXT_TELEGRAM = 0x1F
    IDLE_FILLER = 0x2F
    GLOBAL_READOUT_REQUEST = 0x7F


class SpecialFunctionsBlock(AbstractBlock):
    """Special functions block."""

    DATA_FIELD_SPECIFIERS = [
        0x0F,
    ]

    def __init__(self, function: Function = Function.END_OF_USER_DATA):
        """Initialize special functions block."""
        super().__init__()

        self.function = function

    @property
    def dif(self):
        """Return DIF."""
        return self.function.value

    @classmethod
    def from_stream(cls, dif, stream):
        """Construct SpecialFunctionsBlock from stream."""
        block = cls()
        block.function = Function(dif)

        return block

    def __bytes__(self):
        """Return byte representation of block."""
        return b"%c" % self.dif



"""Abstract definition of MBus frame blocks."""

from abc import ABC, abstractmethod


def information_block(first_byte: int, extension: list[int]) -> bytes:
    """
    Format the data or value information block.

    :param fb: First byte, i.e. DIF or VIF
    :param ext: Extension bytes, i.e. DIFE or VIFE
    """
    if first_byte is None:
        return b""

    if len(extension) == 0:
        # If no extension is present, erase the extension bit
        return b"%c" % (first_byte & 0x7F)

    *_es, _e = extension
    return b"%c%b%c" % (first_byte | 0x80, bytes([_b | 0x80 for _b in _es]), _e & 0x7F)


class AbstractBlock(ABC):
    """Block in variable data MBus frame."""

    DATA_FIELD_SPECIFIERS = []

    def __init__(self):
        """Initialize block."""
        self._dif = 0x00
        self._dife = []

        # Initialize vif and vife variables
        self._vif = None
        self._vife = []

    @property
    def dif(self):
        """Return DIF."""
        return self._dif

    @property
    def dife(self):
        """Return DIFE."""
        return self._dife

    @property
    def vif(self):
        """Return VIF."""
        return self._vif

    @property
    def vife(self):
        """Return VIFE."""
        return self._vife

    @property
    def drh(self):
        """Return the data record header."""
        return b"%b%b" % (
            information_block(self.dif, self.dife),
            information_block(self.vif, self.vife),
        )

    @classmethod
    @abstractmethod
    def from_stream(cls, dif, stream):
        """Construct block from stream."""

    @abstractmethod
    def __bytes__(self):
        """Return byte representation of block."""


class AbstractDataBlock(AbstractBlock):
    """Block representing a generic data value and unit."""

    def __init__(self):
        """Inititalize block."""
        super().__init__()

        self._value = 0

    @property
    def value(self):
        """Return value of data block."""
        return self._value

    @value.setter
    def value(self, new_value):
        """Set value of data block."""
        self._value = new_value

    @classmethod
    def from_stream(cls, dif, stream):
        """Parse data block from stream."""
        block = cls()
        block._dif = dif

        # Read data information field extensions
        _e = dif & 0x80
        while _e:
            dife = ord(stream.read(1))
            block._dife.append(dife)

            _e = dife & 0x80

        # Read value information field
        block._vif = vif = ord(stream.read(1))

        # Read value information field extensions
        _e = vif & 0x80
        while _e:
            vife = ord(stream.read(1))
            block._vife.append(vife)

            _e = vife & 0x80

        return block

    @property
    @abstractmethod
    def data(self) -> bytes:
        """Return byte representation of data."""

    def __bytes__(self) -> bytes:
        """Return byte representation of data block."""
        return b"%b%b" % (self.drh, self.data)



"""Speciality block types."""

from ._abstract_block import AbstractDataBlock


class NoDataBlock(AbstractDataBlock):
    """No data block."""

    DATA_FIELD_SPECIFIERS = [
        0x00,
    ]

    @property
    def data(self):
        """Return byte representation of data."""
        return b""


class SelectionForReadoutBlock(AbstractDataBlock):
    """Selection for readout block."""

    DATA_FIELD_SPECIFIERS = [
        0x08,
    ]

    @property
    def data(self):
        """Return byte representation of data."""
        return b""



"""Common methods and classes."""

import enum


class Byte(enum.Enum):
    """Byte enum type."""

    def __init__(self, byte, mask, info=None):
        """Create Byte enumeration."""
        self._value_ = byte
        self.mask = mask

        if info is not None:
            self.info = info

    @classmethod
    def _missing_(cls, value):
        for item in cls:
            if (value & item.mask) == item.value:
                return item

        raise KeyError(f"No such item {value:02X}")



"""Data block representing an floating point value."""

import struct

from ._abstract_block import AbstractDataBlock


class RealBlock(AbstractDataBlock):
    """Floating point data block."""

    DATA_FIELD_SPECIFIERS = [
        0x05,
    ]

    def __init__(self, value=0):
        """Initialize floating point data block."""
        super().__init__()

        self.value = value

    def __len__(self):
        """Return length of data field."""
        return 4

    @property
    def dif(self):
        """Return DIF."""
        return (self._dif & 0xF0) | 0x05

    @property
    def data(self):
        """Get byte representation of value."""
        return struct.pack("<f", self.value)

    @classmethod
    def from_stream(cls, dif, stream):
        """Construct block from stream."""
        # Create instance and parse DRH
        block = super().from_stream(dif, stream)

        data = stream.read(len(block))
        (block.value,) = struct.unpack("<f", data)

        return block



"""Fixed MBus data structure."""


class FixedData:
    """Fixed data structure."""

    def __init__(self):
        """Initialize fixed data stricture."""
        raise NotImplementedError("Fixed data structure not implemented yet")

    def __bytes__(self):
        """Return byte representation of data structure."""
        return b""

    @classmethod
    def from_bytes(cls, data):
        """Parse data structure from stream."""
        raise NotImplementedError("Parsing of fixed data not implemented yet")



"""BCD data block."""

from ._abstract_block import AbstractDataBlock


def encode_bcd(value, length):
    """
    Encode unsigned integer value as BCD.

    :param value: Integer value to encode
    :param length: Byte length of output
    """
    _div = 100**length
    _bs = []

    assert _div > value > 0, "Value out of bounds"

    while _div > 1:
        _div = _div // 100

        _a, _b = divmod(value // _div, 10)

        _bs.append((_a << 4) | _b)
        value %= _div

    return bytes(_bs)


def decode_bcd(byte_string):
    """
    Decode binary coded decimal.

    :param byte_string: Byte string to decode
    """
    value = 0
    for byte in list(byte_string):
        value *= 100
        value += ((byte & 0xF0) >> 4) * 10 + (byte & 0x0F)

    return value


class BCDBlock(AbstractDataBlock):
    """BCD data block."""

    _DEF = [
        (0x09, 1),
        (0x0A, 2),
        (0x0B, 3),
        (0x0C, 4),
        (0x0E, 6),
    ]

    _DIF = {l: b for (b, l) in _DEF}
    _LENGTH = {b: l for (b, l) in _DEF}

    DATA_FIELD_SPECIFIERS = [b for b, _ in _DEF]

    def __init__(self, value=0, length=1):
        """Create BCD data block."""
        super().__init__()

        self.value = value
        self.length = length

    def __len__(self):
        """Length of data."""
        return self.length

    @property
    def dif(self):
        """Return DIF."""
        return (self._dif & 0xF0) | self._DIF[len(self)]

    @classmethod
    def from_stream(cls, dif, stream):
        """Create BCDBlock from stream."""
        block = super().from_stream(dif, stream)  # type: 'BCDBlock'

        block.length = cls._LENGTH[dif & 0x0F]
        block.value = decode_bcd(stream.read(len(block)))

        return block

    @property
    def data(self):
        """Encode value."""
        return encode_bcd(self.value, len(self))



"""MBus frame."""

from ._frame import Control, ControlInformation, Frame, FrameType
from ._raw_frame import RawMBusFrame

__all__ = [
    "Control",
    "ControlInformation",
    "Frame",
    "FrameType",
    "RawMBusFrame",
]
